@@ -11,9 +11,18 @@ from mwcp.parser import Parser
 from mwcp.file_object import FileObject
 from mwcp.registry import (
     register_entry_points, register_parser_directory, register_parser_package,
-    iter_parsers, get_parser_descriptions, set_default_source, clear_default_source,
+    iter_parsers, get_parser_descriptions, set_default_source,
+    clear as clear_registry,
+    clear_default_source,
     ParserNotFoundError
 )
-from mwcp.reporter import Reporter
-from mwcp.dispatcher import Dispatcher, UnableToParse, UnidentifiedFile
+from mwcp.reporter import Reporter  # DEPRECATED
+from mwcp.runner import Runner
+from mwcp.report import Report
+from mwcp.dispatcher import Dispatcher, UnidentifiedFile
 from mwcp.utils.logutil import setup_logging
+from mwcp.core import run, schema
+from mwcp.exceptions import *
+
+
+__version__ = "3.5.0"
